#!/usr/bin/perl -w

###############################################################
#
#      File: saywhat.perl
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: Produce nonsense to the dsp with
#              festival
#
###############################################################

use FileHandle;
use Getopt::Long;
use Pod::Usage;
use Carp;

# testing hack
use lib qw(.);

use SayWhat::Base;
#use SayWhat::Tk;
use SayWhat::Parser;
use SayWhat::LexWrapper;
use Grammar::ContextFree::Generator;
#use Storable;

GetOptions(\%opts,
	   ## General Options
	   "help!",
	   "man!",
	   "version!",
	   "verbose:i",
	   "debug!",
	   "safe!",
	   ## Mode options
	   "tk!",
	   "pd!",
	   "osc!",
	   ## OSC-mode options
	   "oschost|osch:s",
	   "oscport|oscp=i",
	   "oscbuflen|oscb=i",
	   "oscmaxlen|oscm=i",
	   "oscspeed|oscs=i",
	   "osccoeff|oscc=i",
	   ## Pd-mode Options
	   "pdhost|pdh:s",
	   "pdsend|pds:i",
	   "pdreceive|pdr:i",
	   "pdfifo:s",
	   ## Grammar Options
	   "wbound=s",
	   "start=s",
	   "maxexpansions=i",
	   ## Synthesizer Options
	   "stype=s",
	   "rate=i",
	   "dummy!",
	   ## Festival Options
	   "voice=s",
	   "style=s",
	   "fserver=s",
	   "fport=i",
	  );

#--------------------------------------------------------------
# Process Options: help, version
#--------------------------------------------------------------
if (!defined($opts{verbose}) || $opts{verbose} >= 0) {
  print STDERR "\nSayWhat version $SayWhat::Base::VERSION by Bryan Jurish <moocow\@ling.uni-potsdam.de>\n\n";
  exit(0) if ($opts{version});
}
pod2usage({
	   -exitval => 1,
	   -verbose => 2,
	  }) if ($opts{man});
pod2usage({
	   -exitval => 1,
	   -verbose => 0,
	  }) if ($opts{help});


#--------------------------------------------------------------
# Process Options: defaults
#--------------------------------------------------------------
## General
$opts{verbose} = 1 unless (defined($opts{verbose}));
$opts{debug} = 0 unless (defined($opts{debug}));
$opts{dummy} = 0 unless (defined($opts{dummy}));
#$opts{safe} = 0 unless (defined($opts{safe}));

## Mode
$opts{tk} = 0 unless (defined($opts{tk}));
$opts{pd} = 0 unless (defined($opts{pd}));
$opts{osc} = 0 unless (defined($opts{osc}));

##-- OSC
$opts{oschost} = 0 if (!defined($opts{oschost}));
$opts{oscport} = 5151 if (!defined($opts{oscport}));
$opts{oscbuflen} = 1024 if (!defined($opts{oscbuflen}));
$opts{oscmaxlen} = 1024 if (!defined($opts{oscmaxlen}));
$opts{oscspeed} = 10 if (!defined($opts{oscspeed}));
$opts{osccoeff} = 100 if (!defined($opts{osccoeff}));

##-- Pd
#$opts{pdfifo} = '/tmp/pd-saywhat.fifo' unless (defined($opts{pdfifo}));
$opts{pdfifo} = undef if (!defined($opts{pdfifo}));
$opts{pdhost} = 'localhost' if (!defined($opts{pdhost}));
$opts{pdreceive} = 42420 if (!defined($opts{pdreceive}));
$opts{pdsend} = 42421 if (!defined($opts{pdsend}));

##-- Grammar
$opts{wbound} = ' ' unless (defined($opts{wbound}));
$opts{start} = 'S' unless (defined($opts{start}));
#$opts{maxexpansions} = undef if (!defined($opts{maxexpansions}));

## Synthesizer
$opts{stype} = 'Dummy' unless (defined($opts{stype}));
#$opts{stype} = 'rsynth' unless (defined($opts{stype}));
$opts{rate} = $opts{pd} ? 44100 : 8000 unless (defined($opts{rate}));

## Festival
$opts{fserver} = 'localhost' unless (defined($opts{fserver}));
$opts{fport} = 1314 unless (defined($opts{fport}));


#--------------------------------------------------------------
# Prepare Synthesizer
#--------------------------------------------------------------
if ($opts{stype} =~ /^[Ff]estival$/i) {
  $opts{stype} = 'SayWhat::Synthesizer::Festival';
} elsif ($opts{stype} =~ /^[Rr]synth$/i) {
  $opts{stype} = 'SayWhat::Synthesizer::Rsynth';
} elsif ($opts{stype} =~ /^[Dd]ummy$/i) {
  ;
} else {
  carp("$0: unknown synthesizer-type '$opts{stype}' defaults to 'Dummy'\n");
  #$opts{sytpe} = 'SayWhat::Synthesizer::Festival';
  #$opts{sytpe} = 'SayWhat::Synthesizer::Rsynth';
  $opts{sytpe} = 'SayWhat::Synthesizer::Dummy';
}

## SayWhat::Synthesizer:* args
%synthargs = ();
if ($opts{stype} eq 'SayWhat::Synthesizer::Festival') {
  %synthargs = (host=>$opts{fserver}, port=>$opts{fport});
} else {
  %synthargs = ();
}

$opts{safe} = 0 if (!$opts{safe});
## common synth arguments
@synthargs{qw(dummy verbose samp_rate safe)} =
  @opts{qw(dummy verbose rate safe)};


#--------------------------------------------------------------
# Set up base object
#--------------------------------------------------------------
%swargs = (
	   wbound => $opts{wbound},
	   verbose => $opts{verbose},
	   dummy => $opts{dummy},
	   synthtype => $opts{stype},
	   synthargs => \%synthargs,
	   safe => $opts{safe},
	   maxexpansions => $opts{maxexpansions},
	  );

# default class
#$class = 'SayWhat';
$class = 'SayWhat::Base';

if ($opts{tk}) {
  #-------------------------------------------
  # Tk mode
  #-------------------------------------------
  $class = 'SayWhat::Tk';
  eval "use $class;";
  die ($@) if ($@);
}
if ($opts{pd}) {
  #-------------------------------------------
  # PD mode
  #-------------------------------------------
  eval "use SayWhat::PD;";
  die ($@) if ($@);
  eval "use Pd::Socket;";
  die ($@) if ($@);

  $swargs{mode} = 'raw';
  $swargs{pdhost} = $opts{pdhost};
  $swargs{pdfifo} = $opts{pdfifo};
  $swargs{pdsend} = $opts{pdsend};
  $swargs{pd} = 1;
  $opts{safe} = 1 if (!defined($opts{safe}));
  $class = $opts{tk} ? 'SayWhat::TkPD' : 'SayWhat::PD';
}
if ($opts{osc}) {
  #-------------------------------------------
  # OSC mode
  #-------------------------------------------
  require OSC;
  require SayWhat::OSC;

  $swargs{osc} = 1;
  $swargs{oschost} = $opts{oschost};
  $swargs{oscport} = $opts{oscport};
  $swargs{oscbuflen} = $opts{oscbuflen};
  $swargs{oscmaxlen} = $opts{oscmaxlen};
  $swargs{oscspeed} = $opts{oscspeed};
  $swargs{oscspeedcoeff} = $opts{osccoeff};
  $opts{safe} = 1 if (!defined($opts{safe}));
  $class = 'SayWhat::OSC';
}

$sw = $class->new(%swargs);


#--------------------------------------------------------------
# Set up initial festival state
#--------------------------------------------------------------
#if (defined($opts{voice})) { $sw->set_voice($opts{voice}); }
#if (defined($opts{style})) { $sw->set_style($opts{style}); }

#--------------------------------------------------------------
# Set up our lexer tokens
#--------------------------------------------------------------
@tokens =
  (
   ## perl-parsing
   'INITIAL,perlblock:LBRACE', '{',
   'INITIAL,perlblock:RBRACE', '}',
   'perlblock:PERL_CODE', '[^{}]*|\n',

   ## Double-quoted strings
   Parse::Token::Quoted->new(
			     Name => 'DQ_STRING',
			     Escape => '\\',
			     Quote => '"',
			    ),
   ## Single-quoted strings
   Parse::Token::Quoted->new(
			     Name => 'SQ_STRING',
			     Escape => '\\',
			     Quote => qq!\'!,
			    ),
   ## C-style comments
   Parse::Token::Delimited->new(
				Name => 'C_COMMENT',
				Start => '/\*',
				End => '\*/',
			       ),
   ## --- Perl-style comments
   'P_COMMENT',  '\#.*',

   ## --- command-tokens

   ## new v0.04
   qw(
      SYNTH_TYPE   synth-type
      SYNTH_CONF   synth-config
     ),

   qw(HELP         help
      LOAD_GRAMMAR load-grammar
      RETR_GRAMMAR retr-grammar
      LOAD_RULES   load-rules
      LOAD_SCRIPT  load
      SAVE_SCRIPT  save-script
      STORE_GRAMMAR store-grammar
      FORGET_GRAMMAR forget-grammar
      SET_WBOUND   set-wbound
      EVAL_PERL    perl
      LL_UTTS      (?:ll)|(?:list-long)
      LS_UTTS      (?:ls)|(?:list)|(?:list-short)
      ADD_UTT      add
      RENAME_UTT   rename
      DEL_UTT      del
      SYNTH        synth
      SAVE_UTT     save
      SAY          say
      GENERATE     (?:gen(?:erate)?\b)
      GENERATE_SAY (?:gens(?:ay)?\b)
      GENERATE_TO  (?:gen(?:erate)?2\b)
      MAX_EXPANSIONS max-expansions
     ),
   qw(BYE          (?:bye)|(?:exit)|(?:quit)),
   ## --- obsolete!
   #qw(
   #   SET_VOICE    set-voice
   #   SET_STYLE    set-style
   #   SEND_FESTIVAL festival
   #  ),

   ## --- miscellaneous punctuation
   qw(EQUAL =),
   'COMMA', ',',

   ## --- end-of-command-tokens
   qw(
      SEMICOLON ;
      NEWLINE \n
     ),

   ## --- symbols
   qw(SYMBOL [\w\d:-]+),

   qw(LEXERROR .*), sub {
     $lexwrapper->yywarn(qq!lexer error: can\'t analyze "$_[1]"!);
   }

  );


#--------------------------------------------------------------
# Set up lexer, lex wrappter, & parser
#--------------------------------------------------------------
$lexer = Parse::Lex->new(@tokens);
$lexer->exclusive('perlblock' => 1);
#$lexer = Parse::Lex->new();
#$lexer->exclusive(perlblock => 1);
#$lexer->configure(Tokens => \@tokens);

# --- ignore backslashed newline-escapes
$lexer->skip("(?:[ \t]*\\\\\n[ \t]*)|(?:[ \t]+)");

$lexwrapper = LexWrapper->new(lexer => $lexer, 
			      ignore => ["COMMENT\$"]);

$parser = SayWhat::Parser->new();
$parser->YYData->{LexWrapper} = $lexwrapper;
$parser->YYData->{SayWhat} = $sw;


#--------------------------------------------------------------
# Help sub
#--------------------------------------------------------------
@helpinfo =
  (
   'help [<regex>]' , 'Show help for commands matching <regex> (a string)',
   'load <file>', 'Load a SayWhat script file from <file> (a string)',
   'save-script <file>', 'Save utterance-information to a SayWhat script file <file> (a string)',
   'load-grammar <file>' , 'Load grammar from <file> (a string), autodetect',
   'retr-grammar <file>' , 'Load compiled grammar from <file> (a string)',
   'load-rules <file>' , 'Load grammar from ASCII file <file> (a string)',
   'store-grammar <file>' , 'Store current (compiled) grammar to <file> (a string)',
   'forget-grammar' , 'Clear current grammar',
   'set-wbound <string>' , 'Set the current word-boundary to <string>',
   'generate [<sym>]' , 'Generate an utterance derived from the nonterminal <sym> (defaults to `S\')',
   'gensay [<sym>]' , 'Generate and pronounce an utterance derived from the nonterminal <sym> (defaults to `S\')',
   'generate2 <id> [<sym>]' , 'Generate utterance <id> from the nonterminal <sym> (defaults to `S\')',
   'perl <perl-block>' , 'Evaluate <perl-block> in package main [caveat nestor]',
   'ls [<regexp>]' , 'List utterances matching <regexp> (short format)',
   'll [<regexp>]' , 'List utterances matching <regexp> (long format)',
   'add <string> [<tree> [<id>]]' , 'Add a new utterance <string>, with optional <tree> (string) and id <id>',
   'rename <old> <new>', 'Rename the utterance identified by <old> to <new>.',
   'del <sym>' , 'Delete the utterance identified by <sym> (a symbol)',
   'synth <sym>' , 'Synthesize the utterance identified by <sym> (a symbol)',
   'save <sym> [<file> [<type>]]' , 'Save the waveform of the utterance identifiied by <sym> to <file>.  <file> defaults to "<sym>.wav.  <type> defaults to "riff" (wav) format."',
   'say <sym>' , 'Pronounce the utterance identifiied by <sym> (a symbol)',
   ## new v0.04
   'synth-type [<type>]', 'Get/set synthesizer type',
   'synth-config [<id>]', 'Get synthsizer configuration [for key <id>]',
   'synth-config <id>=<val> , ...', 'Set synthesizer configuration',
   'max-expansions N , ...', 'Set max grammar expasions',

   ## away wi' ye
   'bye' , 'Exit the SayWhat interpreter',

   ## obsolete
   #'set-voice <voice>' , 'Set the current Festival voice to <voice> (a symbol)',
   #'set-style <id>' ,  'Set the current Festival style to <style> (a symbol)',
   #'festival <scheme-form>' , 'Send <scheme-form> (a string) to the Festival server',
  );

sub help {
  my $regex = shift;
  my ($oldtop,$oldfmt,$oldformfeed) = ($^,$~,$^L);
  ($^,$~,$^L) = ('HELP_TOP', 'HELP', '');
  local ($cmd,$descr);
  my ($i);
  print "   COMMAND                     DESCRIPTION\n";
  for ($i = 0; $i <= $#helpinfo; $i += 2) {
    ($cmd,$descr) = @helpinfo[$i..$i+1];
    next unless (!defined($regex) || $cmd =~ $regex || $descr =~ $regex);
    write;
  }
  ($^,$~,$^L) = ($oldtop,$oldfmt,$oldformfeed);
}
format HELP_TOP =
.
format HELP =
   ^<<<<<<<<<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
   $cmd,                      $descr
~~ ^<<<<<<<<<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
   $cmd,                      $descr

.


#--------------------------------------------------------------
# Signals
#--------------------------------------------------------------
$SIG{KILL} = \&die_gracefully;
$SIG{HUP} = \&die_gracefully;
$SIG{INT} = \&die_gracefully;
$SIG{QUIT} = \&die_gracefully;


#--------------------------------------------------------------
# ...do some parsing
#--------------------------------------------------------------

#-------------------------------------------
# Parse initial files
#-------------------------------------------
if (@ARGV) {
  while ($filename = shift(@ARGV)) {
    open(INFILE, "<$filename") or
      die("open failed for file `$filename': $!");
    #$lexer->from(*INFILE);
    $lexwrapper->from(\*INFILE);
    $lexwrapper->filename($filename);
    $sw->doprompt($lexwrapper);
    doparse();
    close(INFILE);
  }
  exit(0) unless ($opts{tk} || $opts{pd} || $opts{osc});
}

if ($opts{tk}) {
  #-------------------------------------------
  # Tk mode
  #-------------------------------------------
  $SayWhat::Base::Prompt = ''; # no prompts on read-from-file
  Tk::MainLoop();
  exit 0;
}
elsif ($opts{pd}) {
  #-------------------------------------------
  # PD mode
  #-------------------------------------------
  if ($opts{debug}) {
    $netrec = IO::File->new("<&STDIN");
    #$lexer->from(*$netrec);
    $lexwrapper->from(\*$netrec);
    $lexwrapper->filename("(Pd DEBUG)");
    $sw->doprompt($lexwrapper);
    doparse();
  } else {
    $netrec = Pd::Socket::Receive::TCP->new(LocalAddr=>0,
					    LocalPort=>$opts{pdreceive},
					    Listen=>1)
      or die("$0: open failed for pipe from 'pdreceive $opts{pdreceive}': $!");
    while (my $client = $netrec->accept()) {
      #$lexer->from(*$client);
      $lexwrapper->from(\*$client);
      $lexwrapper->filename("(pdreceive $opts{pdreceive})");
      $sw->doprompt($lexwrapper);
      doparse();
    }
  }
}
elsif ($opts{osc}) {
  #-------------------------------------------
  # OSC mode
  #-------------------------------------------
  $lexwrapper->file("(OSC::Dump on $sw->{oschost}:$sw->{oscport})");
  $sw->server(sub {
		my $cmd = shift;
		$lexwrapper->from($cmd."\n");
		doparse();
	      });
}
else {
  #-------------------------------------------
  # Command-line mode
  #-------------------------------------------
  #$lexer->from(*STDIN);
  $lexwrapper->from(*STDIN);
  $lexwrapper->filename('(STDIN)');
  $sw->doprompt($lexwrapper);
  doparse();
}

sub doparse_debug {
  $lexwrapper->dummylex();
}

sub doparse {
  $parser->YYParse(
		   yylex => $lexwrapper->_yylex,
		   yyerror => $lexwrapper->_yyerror,
		   #yydebug => 0x01, # lexer debug
		   #yydebug => 0x02, # state info
		   #yydebug => 0x04,  # driver actions (shift/reduce/etc.)
		   #yydebug => 0x08,   # stack dump
		   #yydebug => 0x10,  # Error recovery trace
		   #yydebug => 0x01 | 0x02 | 0x04 | 0x08,
		  );
}


sub die_gracefully {
  my $sig = shift;
  print STDERR "$0: caught signal SIG$sig -- bailing out.\n";
  if ($opts{pd}) {
    $netrec->close() if ($netrec);
    $sw->{pdsendfh}->close() if (defined($sw->{pdsendfh}));
    unlink($opts{pdfifo}) if defined($opts{pdfifo});
  }
}

__END__

#-----------------------------------------------------------------------
# Pods
#-----------------------------------------------------------------------
=pod

=head1 NAME

saywhat.perl -- a grammatically correct nonsense generator.

=head1 SYNOPSIS

 saywhat.perl OPTIONS SCRIPT_FILE(s)

 General Options:
   -help
   -man
   -verbose LEVEL
   -safe
   -debug

 Mode Options:
   -tk
   -pd
   -osc

 Pd-Mode Options:
   -pdhost HOST    (=localhost)
   -pdsend PORT    (=42421)
   -pdreceive PORT (=42420)
   -pdfifo FIFO

 OSC-Mode Options:
   -oschost HOST   (=0)
   -oscport PORT   (=5151)
   -oscbuflen LEN  (=1024)
   -oscmaxlen LEN  (=1024)
   -oscspeed  MS   (=10)
   -osccoeff  MS   (=100)

 Grammar Options:
   -wbound  STRING
   -start   START
   -maxexpansions N

 Synthesizer Options:
   -stype SYNTH_TYPE
   -rate  SAMP_RATE
   -dummy

 Festival Options:
   -fserver HOST
   -fport   PORT
   -voice   VOICE
   -style   STYLE

=cut

#-----------------------------------------------------------------------
# Options & arguments
#-----------------------------------------------------------------------
=pod

=head1 OPTIONS AND ARGUMENTS

=cut

#-----------------------------------------------------------------------
# General Options
#-----------------------------------------------------------------------
=pod

=head2 General Options

=over 4

=item * -help

Display a brief help message.

=item * -man

Display a longer help message.

=item * -verbose LEVEL

Set verbosity level.  Known levels are:

 -1 : be silent.
  0 : print only generated text to stdout and errors/warnings to stderr.
  1 : print standard command-line messages


=item * -safe , -nosafe

Enable / disable "safe" mode.  In "safe" mode, the "perl {}"
command is disabled.  Useful if running as a server.

Default: disabled in command-line and Tk modes, enabled
in pd and osc modes.


=item * -debug

Display debugging information.

=back

#-----------------------------------------------------------------------
# Mode Options
#-----------------------------------------------------------------------
=pod

=head2 Mode Options

=over 4

=item * -tk

Run in perl/Tk mode.

=item * -pd

Run in PD ("Pure Data") server/client mode --
'say' outputs raw linear 44.1KHz audio data to PDFIFO,
input is read from 'pdreceive PORT_R'.

=item * -osc

Run in OSC ("Open Sound Control") mode -- UDP requests are
read from OSCPORT.  See SayWhat::OSC for details.

=back

=cut

#-----------------------------------------------------------------------
# Pd-Mode Options
#-----------------------------------------------------------------------
=pod

=head2 Pd-Mode Options

=over 4

=item * -pdhost HOST

Host on which Pd is running, for pd-mode.
Default=localhost.

=item * -pdreceive PORT_R

Port on which to listen for messages from PD in pd-mode.
Default=42420.

=item * -pdsend PORT_S

Port on which to send messages to PD in pd-mode.
Default=42421.

=item * -pdfifo FIFO

Fifo (named pipe) to which synthesizer 'say' output should be
written in PD mode.
Default=(none).

=back

=cut


#-----------------------------------------------------------------------
# OSC-Mode Options
#-----------------------------------------------------------------------
=pod

=head2 OSC-Mode Options

=over 4

=item * -oschost HOST

Host on which to bind OSC server socket.
Default=0 (any).

=item * -oscport OPORT_R

Port on which to listen for messages from in OSC-mode.
Default=5151.

=item * -oscbuflen LEN

Set maximum length for incoming OSC messages.
Default=1024.

=item * -oscmaxlen LEN

Set maximum total length for outgoing OSC strings.
Tokens will be sent individually, but this parameter
applies to total string-length.
Default=1024.

=item * -oscspeed MS

Set minimum time between OSC output messages.
Default=10.

=item * -osccoeff MS

Set string-length coefficient for delay between OSC messages.
Default=10.

=back

=cut


#-----------------------------------------------------------------------
# Grammar Options
#-----------------------------------------------------------------------
=pod

=head2 Grammar Options

=over 4

=item * -wbound STRING

Set initial word-boundary for grammar.  Default=' '.

=item * -start START

Set initial grammar start-symbol to START.  Default='S'.

=item * -maxexpansions N

Set maximum number of rule-expansions, to avoid infinite loops.

=back


#-----------------------------------------------------------------------
# Synthesizer Options
#-----------------------------------------------------------------------
=pod

=head2 Synthesizer Options

=over 4

=item * -stype SYNTH_TYPE

Set initial synthesizer type.

=item * -rate SAMP_RATE

Set initial sampling rate for synthesizer output.

=item * -dummy

Do not synthesize anything.

=back

#-----------------------------------------------------------------------
# Festival Options
#-----------------------------------------------------------------------
=pod

=head2 Festival Options

Options for festival-mode synthesis.

=over 4

=item * -fserver HOST

Use festival server on HOST  (default=localhost).

=item * -fport PORT

Use festival port PORT (default=1314).

=item * -voice VOICE

Set the initial voice to VOICE.

=item * -style STYLE

Set the initial style to STYLE.

=back

=cut

#-----------------------------------------------------------------------
# Footer
#-----------------------------------------------------------------------
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

"rsynth" speech synthesizer and perl/Tk by Nick Ing-Simmons.

"festival" speech synthesis system by Alan W Black, Richard
Caley and Paul Taylor and others.

"Pure Data" realtime audio processing software by
Miller Puckette and others.

=head1 AUTHOR

Bryan Jurish E<lt>moocow@ling.uni-potsdam.deE<gt>

=head1 SEE ALSO

perl(1).
say(1).
festival(1).
pd(1).

=cut

